@@ -11,11 +11,11 @@
 //!
 //! ```toml
 //! [[bin]]
-//! path = "src/main.rs"
+//! name = "my-sop"
 //! required-features = ["cli"]
 //!
 //! [[bin]]
-//! path = "src/mainv.rs"
+//! name = "my-sopv"
 //! required-features = ["cliv"]
 //!
 //! [features]
@@ -23,12 +23,34 @@
 //! cliv = ["sop/cliv"]
 //! ```
 //!
-//! And create `src/main.rs` and `src/mainv.rs`, both containing
-//! something along the lines of:
+//! And create `src/bin/my-sop.rs` and `src/bin/my-sopv.rs`,
+//! containing, respectively, something along the lines of:
 //!
 //! ```rust,ignore
 //! fn main() {
-//!     sop::cli::main(&MySOPImplementation::default());
+//!     // Be clever and inspect argv[0].
+//!     let argv0 = std::env::args_os().next();
+//!     let variant = if argv0
+//!         .map(|s| s.to_string_lossy().contains("my-sopv"))
+//!         .unwrap_or(false)
+//!     {
+//!         // Restrict to the verification subset.
+//!         sop::cli::Variant::Verification
+//!     } else {
+//!         sop::cli::Variant::Full
+//!     };
+//!
+//!     sop::cli::main(&mut MySOPImplementation::default(),
+//!                    sop::cli::Variant::Verification);
+//! }
+//! ```
+//!
+//! And for the verification subset:
+//!
+//! ```rust,ignore
+//! fn main() {
+//!     sop::cli::main(&mut MySOPImplementation::default(),
+//!                    sop::cli::Variant::Verification);
 //! }
 //! ```
 //!
@@ -36,7 +58,7 @@
 //! verification subset of SOP, you can also build both binaries from
 //! the same source.
 //!
-//! ## Generating shell completions
+//! ## Generating shell completions and manual pages
 //!
 //! To create shell completions, add this snippet to `Cargo.toml`:
 //!
@@ -48,15 +70,57 @@
 //! And create `build.rs` along the lines of:
 //!
 //! ```rust,no_run
+//! use std::{
+//!     env,
+//!     error::Error,
+//!     fs,
+//!     path::PathBuf,
+//! };
+//!
+//! use sop::cli::Variant;
+//!
+//! #[path = "src/version.rs"]
+//! mod version;
+//!
 //! fn main() {
-//!     #[allow(dead_code)]
-//!     let outdir = std::env::var_os("CARGO_TARGET_DIR")
-//!         .or(std::env::var_os("OUT_DIR"))
-//!         .expect("cargo to set OUT_DIR");
 //!     #[cfg(feature = "cli")]
-//!     sop::cli::write_shell_completions("sqop", &outdir).unwrap();
+//!     cli("sqop", Variant::Full).unwrap();
+//!
 //!     #[cfg(feature = "cliv")]
-//!     sop::cli::write_shell_completions("sqopv", &outdir).unwrap();
+//!     cli("sqopv", Variant::Verification).unwrap();
+//! }
+//!
+//! /// Writes shell completions and man pages.
+//! fn cli(name: &'static str, variant: Variant) -> Result<(), Box<dyn Error>> {
+//!     let version = version::Version::with_name(name);
+//!
+//!     sop::cli::write_shell_completions2(
+//!         variant, name, asset_out_dir("shell-completions")?)?;
+//!     sop::cli::write_man_pages(
+//!         variant, &version, "Sequoia PGP",
+//!         asset_out_dir("man-pages")?)?;
+//!
+//!     Ok(())
+//! }
+//!
+//! /// Variable name to control the asset out directory with.
+//! const ASSET_OUT_DIR: &str = "ASSET_OUT_DIR";
+//!
+//! /// Returns the directory to write the given assets to.
+//! #[allow(dead_code)]
+//! fn asset_out_dir(asset: &str) -> Result<PathBuf, Box<dyn Error>> {
+//!     println!("cargo:rerun-if-env-changed={}", ASSET_OUT_DIR);
+//!     let outdir: PathBuf =
+//!         env::var_os(ASSET_OUT_DIR).unwrap_or_else(
+//!             || env::var_os("OUT_DIR").expect("OUT_DIR not set")).into();
+//!     if outdir.exists() && ! outdir.is_dir() {
+//!         return Err(
+//!             format!("{}={:?} is not a directory", ASSET_OUT_DIR, outdir).into());
+//!     }
+//!
+//!     let path = outdir.join(asset);
+//!     fs::create_dir_all(&path)?;
+//!     Ok(path)
 //! }
 //! ```
 //!
@@ -92,19 +156,29 @@ use crate::{
     SignAs,
 };
 
+/// The last meaningful update of the command line interface.
+///
+/// This will be used as-is in the generated manual pages.
+pub const LAST_UPDATE_TIMESTAMP: &'static str =
+// !!! Update this with every release changing the interface. !!!
+//
+// date +"%B %Y"
+    "February 2025";
+
 #[derive(Debug, Parser)]
 #[clap(about = "An implementation of the Stateless OpenPGP Command Line Interface")]
 #[clap(disable_version_flag(true))]
-struct Cli {
+struct Cli<O: clap::Subcommand> {
     /// Emit verbose output for debugging.
     #[clap(long = "debug", global = true)]
     debug: bool,
 
     #[clap(subcommand)]
-    operation: Operation,
+    operation: O,
 }
 
 /// Variant of the SOP specification.
+#[derive(Clone, Copy)]
 pub enum Variant {
     /// The full functionality.
     Full,
@@ -113,7 +187,7 @@ pub enum Variant {
     Verification,
 }
 
-impl Cli {
+impl<O: clap::Subcommand> Cli<O> {
     fn with_version<'v, V>(v: &V, variant: Variant)
                        -> Result<clap::Command>
     where
@@ -137,16 +211,16 @@ impl Cli {
             )
         };
 
-        Ok(Cli::command()
+        Ok(Cli::<O>::command()
            .name(Box::leak(name.into_boxed_str()) as &str)
            .about(Box::leak(about.into_boxed_str()) as &str)
            .version(Box::leak(version.into_boxed_str()) as &str))
     }
 }
 
-
+/// All the operations of the verification subset of SOP.
 #[derive(Debug, Parser)]
-enum Operation {
+enum VerificationSubset {
     /// Prints version information.
     ///
     /// Invoked without arguments, returns name and version of the SOP
@@ -175,6 +249,48 @@ enum Operation {
         #[clap(long, conflicts_with("backend"), conflicts_with("extended"), conflicts_with("sop_spec"))]
         sopv: bool,
     },
+
+    /// Verifies Detached Signatures.
+    #[clap(display_order = 500)]
+    #[cfg(any(feature = "cli", feature = "cliv"))]
+    Verify {
+        /// Consider signatures before this date invalid.
+        #[clap(long)]
+        not_before: Option<NotBefore>,
+        /// Consider signatures after this date invalid.
+        #[clap(long)]
+        not_after: Option<NotAfter>,
+        /// Signatures to verify.
+        signatures: String,
+        /// Certs for verification.
+        certs: Vec<String>,
+    },
+
+    /// Verifies Inline-Signed Messages.
+    #[clap(display_order = 751)]
+    #[cfg(any(feature = "cli", feature = "cliv"))]
+    InlineVerify {
+        /// Consider signatures before this date invalid.
+        #[clap(long)]
+        not_before: Option<NotBefore>,
+        /// Consider signatures after this date invalid.
+        #[clap(long)]
+        not_after: Option<NotAfter>,
+        /// Write verification result here.
+        #[clap(long)]
+        verifications_out: Option<String>,
+        /// Certs for verification.
+        certs: Vec<String>,
+    },
+}
+
+/// All the operations of the full SOP.
+#[derive(Debug, Parser)]
+enum Operation {
+    ///
+    #[command(flatten)]
+    VerificationSubset(VerificationSubset),
+
     /// Emits a list of profiles supported by the identified
     /// subcommand.
     #[clap(display_order = 150)]
@@ -363,21 +479,6 @@ enum Operation {
         /// Keys for signing.
         keys: Vec<String>,
     },
-    /// Verifies Detached Signatures.
-    #[clap(display_order = 500)]
-    #[cfg(any(feature = "cli", feature = "cliv"))]
-    Verify {
-        /// Consider signatures before this date invalid.
-        #[clap(long)]
-        not_before: Option<NotBefore>,
-        /// Consider signatures after this date invalid.
-        #[clap(long)]
-        not_after: Option<NotAfter>,
-        /// Signatures to verify.
-        signatures: String,
-        /// Certs for verification.
-        certs: Vec<String>,
-    },
     /// Encrypts a Message.
     #[clap(display_order = 600)]
     #[cfg(feature = "cli")]
@@ -461,22 +562,6 @@ enum Operation {
         #[clap(long)]
         signatures_out: String,
     },
-    /// Verifies Inline-Signed Messages.
-    #[clap(display_order = 751)]
-    #[cfg(any(feature = "cli", feature = "cliv"))]
-    InlineVerify {
-        /// Consider signatures before this date invalid.
-        #[clap(long)]
-        not_before: Option<NotBefore>,
-        /// Consider signatures after this date invalid.
-        #[clap(long)]
-        not_after: Option<NotAfter>,
-        /// Write verification result here.
-        #[clap(long)]
-        verifications_out: Option<String>,
-        /// Certs for verification.
-        certs: Vec<String>,
-    },
     /// Creates Inline-Signed Messages.
     #[clap(display_order = 752)]
     #[cfg(feature = "cli")]
@@ -500,11 +585,30 @@ enum Operation {
 }
 
 /// Generates shell completions.
+#[deprecated(note="Use write_shell_completions2.")]
 pub fn write_shell_completions<B, P>(binary_name: B,
                                      out_path: P)
                                      -> io::Result<()>
 where B: AsRef<str>,
       P: AsRef<Path>,
+{
+    let variant = if binary_name.as_ref().ends_with("v") {
+        Variant::Verification
+    } else {
+        Variant::Full
+    };
+
+    write_shell_completions2(variant, binary_name, out_path)
+}
+
+/// Generates shell completions for the given variant.
+pub fn write_shell_completions2<B, P>(variant: Variant,
+                                      binary_name: B,
+                                      out_path: P)
+                                      -> io::Result<()>
+where
+    B: AsRef<str>,
+    P: AsRef<Path>,
 {
     use clap_complete::{generate_to, Shell};
 
@@ -512,13 +616,28 @@ where B: AsRef<str>,
     let out_path = out_path.as_ref();
     std::fs::create_dir_all(&out_path)?;
 
-    let mut clap = Cli::command();
-    for shell in &[Shell::Bash, Shell::Fish, Shell::Zsh, Shell::PowerShell,
-                   Shell::Elvish] {
-        generate_to(*shell, &mut clap, binary_name, out_path)?;
+    const SHELLS: &[Shell] = &[
+        Shell::Bash, Shell::Fish, Shell::Zsh, Shell::PowerShell,
+        Shell::Elvish,
+    ];
+    match variant {
+        Variant::Verification => {
+            let mut clap = Cli::<VerificationSubset>::command();
+            for shell in SHELLS {
+                generate_to(*shell, &mut clap, binary_name, out_path)?;
+            }
+        },
+
+        Variant::Full => {
+            let mut clap = Cli::<Operation>::command();
+            for shell in SHELLS {
+                generate_to(*shell, &mut clap, binary_name, out_path)?;
+            }
+        },
     }
 
-    println!("cargo:warning=shell completions written to {}",
+    println!("cargo:warning={}: shell completions written to {}",
+             binary_name,
              out_path.display());
     Ok(())
 }
@@ -537,15 +656,60 @@ where
     let out_path = out_path.as_ref();
     std::fs::create_dir_all(&out_path)?;
 
-    let cmd = Cli::with_version(version, variant)?
-        .author(Box::leak(author.to_string().into_boxed_str()) as &str);
-    clap_mangen::generate_to(cmd, out_path)?;
+    match variant {
+        Variant::Verification => {
+            let cmd = Cli::<VerificationSubset>::with_version(version, variant)?
+                .author(Box::leak(author.to_string().into_boxed_str()) as &str);
+            generate_to(cmd, LAST_UPDATE_TIMESTAMP, out_path)?;
+        },
 
-    println!("cargo:warning=man pages written to {}",
+        Variant::Full => {
+            let cmd = Cli::<Operation>::with_version(version, variant)?
+                .author(Box::leak(author.to_string().into_boxed_str()) as &str);
+            generate_to(cmd, LAST_UPDATE_TIMESTAMP, out_path)?;
+        },
+    }
+
+    println!("cargo:warning={}: man pages written to {}",
+             version.frontend()?.name,
              out_path.display());
     Ok(())
 }
 
+/// Like clap's version, but modifies the man page's source and date.
+fn generate_to(
+    cmd: clap::Command,
+    date: &str,
+    out_dir: impl AsRef<std::path::Path>,
+) -> Result<(), std::io::Error> {
+    fn generate(
+        cmd: clap::Command,
+        source: &str,
+        date: &str,
+        out_dir: &std::path::Path,
+    ) -> Result<(), std::io::Error> {
+        for cmd in cmd.get_subcommands().filter(|s| !s.is_hide_set()).cloned() {
+            generate(cmd, source, date, out_dir)?;
+        }
+        clap_mangen::Man::new(cmd)
+            .source(source) // Always use the top-level source.
+            .date(date) // Set the last modification timestamp.
+            .generate_to(out_dir)?;
+        Ok(())
+    }
+
+    // Use the top-level source for all man pages.
+    let source = format!(
+        "{} {}",
+        cmd.get_name(),
+        cmd.get_version().unwrap_or_default()
+    );
+
+    let mut cmd = cmd.disable_help_subcommand(true);
+    cmd.build();
+    generate(cmd, &source, date, out_dir.as_ref())
+}
+
 /// Implements the SOP command-line interface.
 pub fn main<'s, S, C, K, Sigs>(sop: &'s mut S, variant: Variant)
                                -> !
@@ -586,7 +750,38 @@ where
     K: Load<'s, S> + Save,
     Sigs: Load<'s, S> + Save,
 {
-    let app = Cli::with_version(sop.version()?.as_ref(), variant)?;
+    // If we are the full version, but are invoked as the verification
+    // subset, we first parse the arguments using the verification
+    // subset parser.
+    #[cfg(feature = "cli")]
+    if let Variant::Verification = variant {
+        let app = Cli::<VerificationSubset>::with_version(
+            sop.version()?.as_ref(), Variant::Verification)?;
+
+        let matches = match app.try_get_matches() {
+            Ok(v) => v,
+            Err(e) => {
+                use clap::error::ErrorKind;
+                return match e.kind() {
+                    ErrorKind::DisplayHelp => {
+                        e.exit()
+                    },
+                    ErrorKind::UnknownArgument =>
+                        Err(anyhow::Error::from(Error::UnsupportedOption)
+                            .context(format!("{}", e))),
+                    _ => Err(e.into()),
+                };
+            },
+        };
+
+        Cli::<VerificationSubset>::from_arg_matches(&matches)?;
+
+        // If we get here, the syntax checked out.  We now just fall
+        // through to the full parser, which accepts a superset of
+        // what we just parsed.
+    }
+
+    let app = Cli::<Operation>::with_version(sop.version()?.as_ref(), variant)?;
 
     let matches = match app.clone().try_get_matches() {
         Ok(v) => v,
@@ -608,7 +803,9 @@ where
     sop.debug(cli.debug);
 
     match cli.operation {
-        Operation::Version { backend, extended, sop_spec, sopv } => {
+        Operation::VerificationSubset(VerificationSubset::Version {
+            backend, extended, sop_spec, sopv,
+        }) => {
             let version = sop.version()?;
             match (backend, extended, sop_spec, sopv) {
                 (false, false, false, false) => {
@@ -804,7 +1001,9 @@ where
         },
 
         #[cfg(any(feature = "cli", feature = "cliv"))]
-        Operation::Verify { not_before, not_after, signatures, certs, } => {
+        Operation::VerificationSubset(VerificationSubset::Verify {
+            not_before, not_after, signatures, certs,
+        }) => {
             let mut op = sop.verify()?;
             if let Some(t) = not_before {
                 op = op.not_before(t.into());
@@ -979,12 +1178,12 @@ where
         },
 
         #[cfg(any(feature = "cli", feature = "cliv"))]
-        Operation::InlineVerify {
+        Operation::VerificationSubset(VerificationSubset::InlineVerify {
             not_before,
             not_after,
             verifications_out,
             certs,
-        } => {
+        }) => {
             let mut op = sop.inline_verify()?;
 
             if let Some(t) = not_before {
@@ -1037,6 +1236,23 @@ fn is_special_designator<S: AsRef<str>>(file: S) -> bool {
     file.as_ref().starts_with("@")
 }
 
+
+/// Checks whether the given descriptor is valid.
+#[cfg(unix)]
+fn check_fd(fd: std::os::fd::RawFd) -> Result<()> {
+    unsafe {
+        // Try to dup(2) it.
+        let dup = libc::dup(fd);
+
+        if dup > 0 {
+            libc::close(dup);
+            Ok(())
+        } else {
+            Err(std::io::Error::last_os_error().into())
+        }
+    }
+}
+
 /// Loads the given (special) file.
 fn load_file<S: AsRef<str>>(file: S)
                             -> Result<(String, Box<dyn io::Read + Send + Sync>)> {
@@ -1058,6 +1274,7 @@ fn load_file<S: AsRef<str>>(file: S)
 
                 let fd: RawFd = f[4..].parse()
                     .map_err(|_| Error::UnsupportedSpecialPrefix)?;
+                check_fd(fd)?;
                 let f = unsafe {
                     std::fs::File::from_raw_fd(fd)
                 };
@@ -1148,6 +1365,7 @@ fn create_file<S: AsRef<str>>(file: S) -> Result<std::fs::File> {
 
                 let fd: RawFd = f[4..].parse()
                     .map_err(|_| Error::UnsupportedSpecialPrefix)?;
+                check_fd(fd)?;
                 let f = unsafe {
                     std::fs::File::from_raw_fd(fd)
                 };
