@@ -88,8 +88,28 @@ pub mod cli;
 
 /// Loads objects like certs and keys.
 pub trait Load<'s, S: SOP<'s>> {
+    /// Loads objects like certs and keys from stdin.
+    fn from_stdin(sop: &'s S) -> Result<Self>
+    where
+        Self: Sized,
+    {
+        Self::from_reader(sop, &mut io::stdin(), Some("/dev/stdin".into()))
+    }
+
+    /// Loads objects like certs and keys from the given file.
+    fn from_file<P>(sop: &'s S, path: P) -> Result<Self>
+    where
+        Self: Sized,
+        P: AsRef<std::path::Path>,
+    {
+        let path = path.as_ref();
+        Self::from_reader(sop, &mut std::fs::File::open(path)?,
+                          Some(path.display().to_string()))
+    }
+
     /// Loads objects like certs and keys from the given reader.
-    fn from_reader(sop: &'s S, source: &mut (dyn io::Read + Send + Sync))
+    fn from_reader(sop: &'s S, source: &mut (dyn io::Read + Send + Sync),
+                   source_name: Option<String>)
                    -> Result<Self>
     where Self: Sized;
 
@@ -98,12 +118,20 @@ pub trait Load<'s, S: SOP<'s>> {
     where
         Self: Sized,
     {
-        Self::from_reader(sop, &mut io::Cursor::new(source))
+        Self::from_reader(sop, &mut io::Cursor::new(source), None)
     }
+
+    /// Returns the source name, if any.
+    fn source_name(&self) -> Option<&str>;
 }
 
 /// Saves objects like certs and keys.
 pub trait Save {
+    /// Writes objects like certs and keys to stdout.
+    fn to_stdout(&self, armored: bool) -> Result<()> {
+        self.to_writer(armored, &mut io::stdout())
+    }
+
     /// Saves objects like certs and keys, writing them to the given writer.
     fn to_writer(&self, armored: bool, sink: &mut (dyn io::Write + Send + Sync))
                  -> Result<()>;
@@ -127,6 +155,14 @@ pub trait SOP<'s>: Sized {
     /// Signatures.
     type Sigs: Load<'s, Self> + Save + plumbing::SopRef<'s, Self>;
 
+    /// Controls debugging.
+    ///
+    /// If enabled, implementations may emit debugging information in
+    /// any way they see fit, most commonly by printing to stderr.
+    fn debug(&mut self, enable: bool) {
+        let _ = enable;
+    }
+
     /// Gets SOP version information.
     ///
     /// The default implementation returns the version of the spec
@@ -134,7 +170,7 @@ pub trait SOP<'s>: Sized {
     /// implementations.  However, implementations may chose to
     /// override this function to return a more nuanced response.
     fn spec_version(&'s self) -> &'static str {
-        "~draft-dkg-openpgp-stateless-cli-10"
+        "~draft-dkg-openpgp-stateless-cli-11"
     }
 
     /// Completeness of the sopv subset
@@ -162,7 +198,7 @@ pub trait SOP<'s>: Sized {
     /// println!("{}", sop.version()?.extended()?);
     /// # Ok(()) }
     /// ```
-    fn version(&'s self) -> Result<Box<dyn Version + 's>>;
+    fn version(&self) -> Result<Box<dyn Version>>;
 
     /// Generates a Secret Key.
     ///
@@ -196,7 +232,7 @@ pub trait SOP<'s>: Sized {
     /// #     Keys: Load<'s, S> + Save,
     /// # {
     /// let alice_secret =
-    ///     Keys::from_reader(sop, &mut File::open("alice.secret")?)?;
+    ///     Keys::from_file(sop, "alice.secret")?;
     ///
     /// let alice_updated_secret = sop.change_key_password()?
     ///     .old_key_password(Password::new_unchecked(b"hunter2".to_vec()))?
@@ -222,7 +258,7 @@ pub trait SOP<'s>: Sized {
     /// #     Keys: Load<'s, S> + Save,
     /// # {
     /// let alice_secret =
-    ///     Keys::from_reader(sop, &mut File::open("alice.secret")?)?;
+    ///     Keys::from_file(sop, "alice.secret")?;
     ///
     /// let alice_revoked = sop.revoke_key()?
     ///     .with_key_password(Password::new_unchecked(b"hunter2".to_vec()))?
@@ -247,7 +283,7 @@ pub trait SOP<'s>: Sized {
     /// #     Keys: Load<'s, S> + Save,
     /// # {
     /// let alice_secret =
-    ///     Keys::from_reader(sop, &mut File::open("alice.secret")?)?;
+    ///     Keys::from_file(sop, "alice.secret")?;
     ///
     /// let alice_public = sop.extract_cert()?
     ///     .keys(&alice_secret)?;
@@ -256,6 +292,56 @@ pub trait SOP<'s>: Sized {
     fn extract_cert(&'s self) -> Result<Box<
             dyn ExtractCert<Self, Self::Certs, Self::Keys> + 's>>;
 
+    /// Keeps a Secret Key Up-To-Date.
+    ///
+    /// This update will "fix" everything that the implementation
+    /// knows how to fix to bring each Transferable Secret Key up to
+    /// reasonable modern practice. Each Transferable Secret Key
+    /// output must be capable of signing, and (unless --signing-only
+    /// is provided) capable of decryption. The primary key of each
+    /// Transferable Secret Key will not be changed in any way that
+    /// affects its fingerprint.
+    ///
+    /// One important aspect of sop update-key is how it handles
+    /// advertisement of support for various OpenPGP capabilities
+    /// (algorithms, mechanisms, etc). All capabilities that the
+    /// implementation knows it does not support, or knows to be weak
+    /// and/or deprecated MUST be removed from the output Transferable
+    /// Secret Keys. This includes unknown/deprecated flags in the
+    /// Features subpacket, and any unknown/deprecated algorithm IDs
+    /// in algorithm preferences subpackets. For example, an
+    /// implementation compliant with [RFC9580] will never emit a
+    /// Transferable Secret Key with a Preferred Hash Preferences
+    /// subpacket that explicitly indicates support for MD5,
+    /// RIPEMD160, or SHA1.
+    fn update_key(&'s self) -> Result<Box<
+            dyn UpdateKey<Self, Self::Certs, Self::Keys> + 's>>;
+
+    /// Merge OpenPGP Certificates.
+    ///
+    /// This can be used, for example, to absorb a third-party
+    /// certification into a certificate, or to update a certificate's
+    /// feature advertisements without losing local annotations.
+    fn merge_certs(&'s self)
+                   -> Result<Box<dyn MergeCerts<Self, Self::Certs> + 's>>;
+
+    /// Certify OpenPGP Certificate User IDs.
+    ///
+    /// With each Transferable Secret Key provided, add a third-party
+    /// certification to the provided certificates, and
+    /// emit the updated OpenPGP certificates.
+    fn certify_userid(&'s self) -> Result<Box<
+            dyn CertifyUserID<Self, Self::Certs, Self::Keys> + 's>>;
+
+    /// Validate a User ID in an OpenPGP Certificate.
+    ///
+    /// Given a set of authority OpenPGP certificates, succeed if and
+    /// only if all provided OpenPGP certificates are correctly bound
+    /// by at least one valid signature from one authority to the User
+    /// ID or email address in question.
+    fn validate_userid(&'s self) -> Result<Box<
+            dyn ValidateUserID<Self, Self::Certs> + 's>>;
+
     /// Creates Detached Signatures.
     ///
     /// Customize the operation using the builder [`Sign`].
@@ -271,7 +357,7 @@ pub trait SOP<'s>: Sized {
     /// #     Keys: Load<'s, S> + Save,
     /// # {
     /// let alice_secret =
-    ///     Keys::from_reader(sop, &mut File::open("alice.secret")?)?;
+    ///     Keys::from_file(sop, "alice.secret")?;
     ///
     /// let (_micalg, sig) = sop.sign()?
     ///     .keys(&alice_secret)?
@@ -297,9 +383,9 @@ pub trait SOP<'s>: Sized {
     /// #     Sigs: Load<'s, S> + Save,
     /// # {
     /// let alice_public =
-    ///     Certs::from_reader(sop, &mut File::open("alice.public")?)?;
+    ///     Certs::from_file(sop, "alice.public")?;
     /// let sig =
-    ///     Sigs::from_reader(sop, &mut File::open("data.asc")?)?;
+    ///     Sigs::from_file(sop, "data.asc")?;
     ///
     /// let verifications = sop.verify()?
     ///     .certs(&alice_public)?
@@ -328,9 +414,9 @@ pub trait SOP<'s>: Sized {
     /// #     Keys: Load<'s, S> + Save,
     /// # {
     /// let alice_secret =
-    ///     Keys::from_reader(sop, &mut File::open("alice.secret")?)?;
+    ///     Keys::from_file(sop, "alice.secret")?;
     /// let bob_public =
-    ///     Certs::from_reader(sop, &mut File::open("bob.public")?)?;
+    ///     Certs::from_file(sop, "bob.public")?;
     ///
     /// let (_session_key, ciphertext) = sop.encrypt()?
     ///     .sign_with_keys(&alice_secret)?
@@ -360,9 +446,9 @@ pub trait SOP<'s>: Sized {
     /// #     Keys: Load<'s, S> + Save,
     /// # {
     /// let alice_public =
-    ///     Certs::from_reader(sop, &mut File::open("alice.public")?)?;
+    ///     Certs::from_file(sop, "alice.public")?;
     /// let bob_secret =
-    ///     Keys::from_reader(sop, &mut File::open("bob.secret")?)?;
+    ///     Keys::from_file(sop, "bob.secret")?;
     ///
     /// let ((_session_key, verifications), plaintext) = sop.decrypt()?
     ///     .verify_with_certs(&alice_public)?
@@ -452,7 +538,7 @@ pub trait SOP<'s>: Sized {
     /// #     Keys: Load<'s, S> + Save,
     /// # {
     /// let alice_public =
-    ///     Certs::from_reader(sop, &mut File::open("alice.public")?)?;
+    ///     Certs::from_file(sop, "alice.public")?;
     ///
     /// let (verifications, data) = sop.inline_verify()?
     ///     .certs(&alice_public)?
@@ -479,7 +565,7 @@ pub trait SOP<'s>: Sized {
     /// #     Keys: Load<'s, S> + Save,
     /// # {
     /// let alice_secret =
-    ///     Keys::from_reader(sop, &mut File::open("alice.secret")?)?;
+    ///     Keys::from_file(sop, "alice.secret")?;
     ///
     /// let (inline_signed_asc) = sop.inline_sign()?
     ///     .keys(&alice_secret)?
