@@ -116,6 +116,93 @@ pub trait ExtractCert<'s, S: SOP<'s>, Certs: Save, Keys: Load<'s, S>> {
     fn keys(self: Box<Self>, keys: &Keys) -> Result<Certs>;
 }
 
+/// Builder for [`SOP::update_key`].
+pub trait UpdateKey<'s, S: SOP<'s>, Certs: Save, Keys: Load<'s, S>> {
+    /// Don't make the updated key encryption-capable if it isn't
+    /// already.
+    fn signing_only(self: Box<Self>) -> Box<dyn UpdateKey<'s, S, Certs, Keys> + 's>;
+
+    /// Don't advertise support for capabilities that aren't already
+    /// advertised by the key.
+    fn no_added_capabilities(self: Box<Self>)
+                             -> Box<dyn UpdateKey<'s, S, Certs, Keys> + 's>;
+
+    /// Supplies a password to unlock the keys with.
+    ///
+    /// All passwords are tried.  If unlocking a key fails, the whole
+    /// operation fails with [`Error::KeyIsProtected`].
+    fn with_key_password(self: Box<Self>, password: Password)
+                         -> Result<Box<dyn UpdateKey<'s, S, Certs, Keys> + 's>>;
+
+    /// Merges updates into the keys.
+    fn merge_updates(self: Box<Self>, updates: &Certs)
+                     -> Result<Box<dyn UpdateKey<'s, S, Certs, Keys> + 's>>;
+
+    /// Updates the keys.
+    fn update(self: Box<Self>, keys: &Keys) -> Result<Keys>;
+}
+
+/// Builder for [`SOP::merge_certs`].
+pub trait MergeCerts<'s, S: SOP<'s>, Certs: Load<'s, S> + Save> {
+    /// Merges updates into the certs.
+    fn merge_updates(self: Box<Self>, updates: &Certs)
+                     -> Result<Box<dyn MergeCerts<'s, S, Certs> + 's>>;
+
+    /// Merges the certs.
+    fn merge(self: Box<Self>, certs: &Certs) -> Result<Certs>;
+}
+
+/// Builder for [`SOP::certify_userid`].
+pub trait CertifyUserID<'s, S: SOP<'s>, Certs: Save, Keys: Load<'s, S>> {
+    /// Certifies the given User ID.
+    fn userid(self: Box<Self>, userid: String)
+              -> Box<dyn CertifyUserID<'s, S, Certs, Keys> + 's>;
+
+    /// Supplies a password to unlock the keys with.
+    ///
+    /// All passwords are tried.  If unlocking a key fails, the whole
+    /// operation fails with [`Error::KeyIsProtected`].
+    fn with_key_password(self: Box<Self>, password: Password)
+                         -> Result<Box<dyn CertifyUserID<'s, S, Certs, Keys> + 's>>;
+
+    /// Don't require self-signatures on the user IDs to be
+    /// certified.
+    fn no_require_self_sig(self: Box<Self>)
+                           -> Box<dyn CertifyUserID<'s, S, Certs, Keys> + 's>;
+
+    /// Create certifications using these keys.
+    fn keys(self: Box<Self>, keys: &Keys)
+            -> Result<Box<dyn CertifyUserID<'s, S, Certs, Keys> + 's>>;
+
+    /// Create the certifications.
+    fn certify(self: Box<Self>, certs: &Certs)-> Result<Certs>;
+}
+
+/// Builder for [`SOP::validate_userid`].
+pub trait ValidateUserID<'s, S: SOP<'s>, Certs: Load<'s, S>> {
+    /// Supplies authority OpenPGP certificates, i.e. trust roots.
+    fn trust_roots(self: Box<Self>, certs: &Certs)
+                   -> Result<Box<dyn ValidateUserID<'s, S, Certs> + 's>>;
+
+    /// Tests that these certificates have the given User ID.
+    fn target_certs(self: Box<Self>, certs: &Certs)
+                    -> Result<Box<dyn ValidateUserID<'s, S, Certs> + 's>>;
+
+    /// Evaluate the validity of the User ID at the specified time,
+    /// not at the current time.
+    fn validate_at(self: Box<Self>, at: SystemTime)
+                   -> Result<Box<dyn ValidateUserID<'s, S, Certs> + 's>>;
+
+    /// Validates an User ID.
+    fn userid(self: Box<Self>, userid: &str) -> Result<()>;
+
+    /// Validates an email address.
+    ///
+    /// Matched only against the e-mail address part of each correctly
+    /// bound User ID.
+    fn email(self: Box<Self>, email: &str) -> Result<()>;
+}
+
 /// Builder for [`SOP::sign`].
 pub trait Sign<'s, S: SOP<'s>, Keys: Load<'s, S>, Sigs: Save> {
     /// Sets signature mode.
@@ -358,30 +445,39 @@ pub trait Ready<T = ()> {
 }
 
 /// A successful signature verification.
-#[derive(Debug, PartialEq, Eq, PartialOrd, Ord, Hash)]
+#[derive(Debug, Hash)]
 pub struct Verification {
     creation_time: SystemTime,
     signing_key_fingerprint: String,
     signing_cert_fingerprint: String,
     signature_mode: SignatureMode,
-    message: Option<String>,
+    description: Description,
+}
+
+/// A successful signature verification.
+#[derive(Debug, Hash)]
+#[derive(serde::Serialize)]
+pub struct Description {
+    #[serde(skip_serializing_if = "Vec::is_empty")]
+    signers: Vec<String>,
+    #[serde(skip_serializing_if = "Option::is_none")]
+    comment: Option<String>,
+    #[serde(skip_serializing_if = "serde_json::Value::is_null")]
+    ext: serde_json::Value,
 }
 
 #[cfg(any(feature = "cli", feature = "cliv"))]
 impl fmt::Display for Verification {
     fn fmt(&self, f: &mut fmt::Formatter) -> fmt::Result {
         write!(f,
-               "{} {} {} {}{}",
+               "{} {} {} {} {}",
                chrono::DateTime::<chrono::Utc>::from(self.creation_time())
                .format("%Y-%m-%dT%H:%M:%SZ"),
                self.signing_key_fingerprint(),
                self.signing_cert_fingerprint(),
                self.signature_mode(),
-               if let Some(m) = self.message() {
-                   format!(" {}", m)
-               } else {
-                   "".into()
-               })
+               serde_json::to_string(&self.description)
+               .unwrap_or_else(|e| e.to_string()))
     }
 }
 
@@ -412,10 +508,28 @@ impl Verification {
             signing_key_fingerprint,
             signing_cert_fingerprint,
             signature_mode,
-            message: message.into().map(Into::into),
+            description: Description {
+                signers: Default::default(),
+                comment: message.into().map(Into::into),
+                ext: Default::default(),
+            },
         })
     }
 
+    /// Adds an issuer name.
+    ///
+    /// This is used by the SOP cli to name the certificates that were
+    /// used to verify the signature with.
+    pub fn add_signer<N: ToString>(&mut self, name: N) {
+        self.description.signers.push(name.to_string());
+    }
+
+    /// Sets arbitrary, implementation-defined data to the
+    /// verification object.
+    pub fn set_extension_info<E: Into<serde_json::Value>>(&mut self, ext: E) {
+        self.description.ext = ext.into();
+    }
+
     /// Returns the signature's creation time.
     pub fn creation_time(&self) -> SystemTime {
         self.creation_time
@@ -438,7 +552,7 @@ impl Verification {
 
     /// Returns a free-form message describing the verification.
     pub fn message(&self) -> Option<&str> {
-        self.message.as_ref().map(AsRef::as_ref)
+        self.description.comment.as_ref().map(AsRef::as_ref)
     }
 }
 
